#!/usr/bin/env python
# encoding: utf-8

"""Produce a translation matrix to translate words from one language to another, using either
a standard nearest neighbour method or a globally corrected neighbour retrieval method [1]_.

This method can be used to augment the existing phrase tables with more candidate translations, or
filter out errors from the translation tables and known dictionaries [2]_. What's more, it also works
for any two sets of named-vectors where there are some paired-guideposts to learn the transformation.

Examples
--------

How to make translation between two set of word-vectors
=======================================================

Initialize two word-vector models

.. sourcecode:: pycon

    >>> from gensim.models import KeyedVectors
    >>> from gensim.test.utils import datapath
    >>>
    >>> model_en = KeyedVectors.load_word2vec_format(datapath("EN.1-10.cbow1_wind5_hs0_neg10_size300_smpl1e-05.txt"))
    >>> model_it = KeyedVectors.load_word2vec_format(datapath("IT.1-10.cbow1_wind5_hs0_neg10_size300_smpl1e-05.txt"))

Define word pairs (that will be used for construction of translation matrix)

.. sourcecode:: pycon

    >>> word_pairs = [
    ...     ("one", "uno"), ("two", "due"), ("three", "tre"), ("four", "quattro"), ("five", "cinque"),
    ...     ("seven", "sette"), ("eight", "otto"),
    ...     ("dog", "cane"), ("pig", "maiale"), ("fish", "cavallo"), ("birds", "uccelli"),
    ...     ("apple", "mela"), ("orange", "arancione"), ("grape", "acino"), ("banana", "banana")
    ... ]

Fit :class:`~gensim.models.translation_matrix.TranslationMatrix`

.. sourcecode:: pycon

    >>> trans_model = TranslationMatrix(model_en, model_it, word_pairs=word_pairs)

Apply model (translate words "dog" and "one")

.. sourcecode:: pycon

    >>> trans_model.translate(["dog", "one"], topn=3)
    OrderedDict([('dog', [u'cane', u'gatto', u'cavallo']), ('one', [u'uno', u'due', u'tre'])])


Save / load model

.. sourcecode:: pycon

    >>> with temporary_file("model_file") as fname:
    ...     trans_model.save(fname)  # save model to file
    ...     loaded_trans_model = TranslationMatrix.load(fname)  # load model


How to make translation between two :class:`~gensim.models.doc2vec.Doc2Vec` models
==================================================================================

Prepare data and models

.. sourcecode:: pycon

    >>> from gensim.test.utils import datapath
    >>> from gensim.test.test_translation_matrix import read_sentiment_docs
    >>> from gensim.models import Doc2Vec
    >>>
    >>> data = read_sentiment_docs(datapath("alldata-id-10.txt"))[:5]
    >>> src_model = Doc2Vec.load(datapath("small_tag_doc_5_iter50"))
    >>> dst_model = Doc2Vec.load(datapath("large_tag_doc_10_iter50"))

Train backward translation

.. sourcecode:: pycon

    >>> model_trans = BackMappingTranslationMatrix(data, src_model, dst_model)
    >>> trans_matrix = model_trans.train(data)


Apply model

.. sourcecode:: pycon

    >>> result = model_trans.infer_vector(dst_model.dv[data[3].tags])


References
----------
.. [1] Dinu, Georgiana, Angeliki Lazaridou, and Marco Baroni. "Improving zero-shot learning by mitigating the
       hubness problem", https://arxiv.org/abs/1412.6568
.. [2] Tomas Mikolov, Ilya Sutskever, Kai Chen, Greg Corrado, and Jeffrey Dean.
       "Distributed Representations of Words and Phrases and their Compositionality", https://arxiv.org/abs/1310.4546

"""

import warnings
from collections import OrderedDict

import numpy as np

from gensim import utils


class Space:
    """An auxiliary class for storing the the words space."""

    def __init__(self, matrix, index2word):
        """
        Parameters
        ----------
        matrix : iterable of numpy.ndarray
            Matrix that contains word-vectors.
        index2word : list of str
            Words which correspond to the `matrix`.

        """
        self.mat = matrix
        self.index2word = index2word

        # build a dict to map word to index
        self.word2index = {}
        for idx, word in enumerate(self.index2word):
            self.word2index[word] = idx

    @classmethod
    def build(cls, lang_vec, lexicon=None):
        """Construct a space class for the lexicon, if it's provided.

        Parameters
        ----------
        lang_vec : :class:`~gensim.models.keyedvectors.KeyedVectors`
            Model from which the vectors will be extracted.
        lexicon : list of str, optional
            Words which contains in the `lang_vec`, if `lexicon = None`, the lexicon is all the lang_vec's word.

        Returns
        -------
        :class:`~gensim.models.translation_matrix.Space`
            Object that stored word-vectors

        """
        # `words` to store all the words
        # `mat` to store the word vector for each word in the 'words' list
        words = []
        mat = []
        if lexicon is not None:
            # if the lexicon is not provided, using all the Keyedvectors's words as default
            for item in lexicon:
                words.append(item)
                mat.append(lang_vec.vectors[lang_vec.get_index(item)])

        else:
            for item in lang_vec.index_to_key:
                words.append(item)
                mat.append(lang_vec.vectors[lang_vec.get_index(item)])

        return Space(mat, words)

    def normalize(self):
        """Normalize the word vectors matrix."""
        self.mat = self.mat / np.sqrt(np.sum(np.square(self.mat), axis=1, keepdims=True))


class TranslationMatrix(utils.SaveLoad):
    """Objects of this class realize the translation matrix which maps the source language to the target language.
    The main methods are:

    We map it to the other language space by computing z = Wx, then return the
    word whose representation is close to z.

    For details on use, see the tutorial notebook [3]_

    Examples
    --------
    .. sourcecode:: pycon

        >>> from gensim.models import KeyedVectors
        >>> from gensim.test.utils import datapath
        >>> en = datapath("EN.1-10.cbow1_wind5_hs0_neg10_size300_smpl1e-05.txt")
        >>> it = datapath("IT.1-10.cbow1_wind5_hs0_neg10_size300_smpl1e-05.txt")
        >>> model_en = KeyedVectors.load_word2vec_format(en)
        >>> model_it = KeyedVectors.load_word2vec_format(it)
        >>>
        >>> word_pairs = [
        ...     ("one", "uno"), ("two", "due"), ("three", "tre"), ("four", "quattro"), ("five", "cinque"),
        ...     ("seven", "sette"), ("eight", "otto"),
        ...     ("dog", "cane"), ("pig", "maiale"), ("fish", "cavallo"), ("birds", "uccelli"),
        ...     ("apple", "mela"), ("orange", "arancione"), ("grape", "acino"), ("banana", "banana")
        ... ]
        >>>
        >>> trans_model = TranslationMatrix(model_en, model_it)
        >>> trans_model.train(word_pairs)
        >>> trans_model.translate(["dog", "one"], topn=3)
        OrderedDict([('dog', [u'cane', u'gatto', u'cavallo']), ('one', [u'uno', u'due', u'tre'])])


    References
    ----------
    .. [3] https://github.com/RaRe-Technologies/gensim/blob/3.2.0/docs/notebooks/translation_matrix.ipynb

    """
    def __init__(self, source_lang_vec, target_lang_vec, word_pairs=None, random_state=None):
        """
        Parameters
        ----------
        source_lang_vec : :class:`~gensim.models.keyedvectors.KeyedVectors`
            Word vectors for source language.
        target_lang_vec : :class:`~gensim.models.keyedvectors.KeyedVectors`
            Word vectors for target language.
        word_pairs : list of (str, str), optional
            Pairs of words that will be used for training.
        random_state : {None, int, array_like}, optional
            Seed for random state.

        """

        self.source_word = None
        self.target_word = None
        self.source_lang_vec = source_lang_vec
        self.target_lang_vec = target_lang_vec

        self.random_state = utils.get_random_state(random_state)
        self.translation_matrix = None
        self.source_space = None
        self.target_space = None

        if word_pairs is not None:
            if len(word_pairs[0]) != 2:
                raise ValueError("Each training data item must contain two different language words.")
            self.train(word_pairs)

    def train(self, word_pairs):
        """Build the translation matrix to map from source space to target space.

        Parameters
        ----------
        word_pairs : list of (str, str), optional
            Pairs of words that will be used for training.

        """
        self.source_word, self.target_word = zip(*word_pairs)

        self.source_space = Space.build(self.source_lang_vec, set(self.source_word))
        self.target_space = Space.build(self.target_lang_vec, set(self.target_word))

        self.source_space.normalize()
        self.target_space.normalize()

        m1 = self.source_space.mat[[self.source_space.word2index[item] for item in self.source_word], :]
        m2 = self.target_space.mat[[self.target_space.word2index[item] for item in self.target_word], :]

        self.translation_matrix = np.linalg.lstsq(m1, m2, -1)[0]

    def save(self, *args, **kwargs):
        """Save the model to a file. Ignores (doesn't store) the `source_space` and `target_space` attributes."""
        kwargs['ignore'] = kwargs.get('ignore', ['source_space', 'target_space'])
        super(TranslationMatrix, self).save(*args, **kwargs)

    def apply_transmat(self, words_space):
        """Map the source word vector to the target word vector using translation matrix.

        Parameters
        ----------
        words_space : :class:`~gensim.models.translation_matrix.Space`
            `Space` object constructed for the words to be translated.

        Returns
        -------
        :class:`~gensim.models.translation_matrix.Space`
            `Space` object constructed for the mapped words.

        """
        return Space(np.dot(words_space.mat, self.translation_matrix), words_space.index2word)

    def translate(self, source_words, topn=5, gc=0, sample_num=None, source_lang_vec=None, target_lang_vec=None):
        """Translate the word from the source language to the target language.

        Parameters
        ----------
        source_words : {str, list of str}
            Single word or a list of words to be translated
        topn : int, optional
            Number of words that will be returned as translation for each `source_words`
        gc : int, optional
            Define translation algorithm, if `gc == 0` - use standard NN retrieval,
            otherwise, use globally corrected neighbour retrieval method (as described in [1]_).
        sample_num : int, optional
            Number of words to sample from the source lexicon, if `gc == 1`, then `sample_num` **must** be provided.
        source_lang_vec : :class:`~gensim.models.keyedvectors.KeyedVectors`, optional
            New source language vectors for translation, by default, used the model's source language vector.
        target_lang_vec : :class:`~gensim.models.keyedvectors.KeyedVectors`, optional
            New target language vectors for translation, by default, used the model's target language vector.

        Returns
        -------
        :class:`collections.OrderedDict`
            Ordered dict where each item is `word`: [`translated_word_1`, `translated_word_2`, ...]

        """
        if isinstance(source_words, str):
            # pass only one word to translate
            source_words = [source_words]

        # If the language word vector not provided by user, use the model's
        # language word vector as default
        if source_lang_vec is None:
            warnings.warn(
                "The parameter source_lang_vec isn't specified, "
                "use the model's source language word vector as default."
            )
            source_lang_vec = self.source_lang_vec

        if target_lang_vec is None:
            warnings.warn(
                "The parameter target_lang_vec isn't specified, "
                "use the model's target language word vector as default."
            )
            target_lang_vec = self.target_lang_vec

        # If additional is provided, bootstrapping vocabulary from the source language word vector model.
        if gc:
            if sample_num is None:
                raise RuntimeError(
                    "When using the globally corrected neighbour retrieval method, "
                    "the `sample_num` parameter(i.e. the number of words sampled from source space) must be provided."
                )
            lexicon = set(source_lang_vec.index_to_key)
            addition = min(sample_num, len(lexicon) - len(source_words))
            lexicon = self.random_state.choice(list(lexicon.difference(source_words)), addition)
            source_space = Space.build(source_lang_vec, set(source_words).union(set(lexicon)))
        else:
            source_space = Space.build(source_lang_vec, source_words)
        target_space = Space.build(target_lang_vec, )

        # Normalize the source vector and target vector
        source_space.normalize()
        target_space.normalize()

        # Map the source language to the target language
        mapped_source_space = self.apply_transmat(source_space)

        # Use the cosine similarity metric
        sim_matrix = -np.dot(target_space.mat, mapped_source_space.mat.T)

        # If `gc=1`, using corrected retrieval method
        if gc:
            srtd_idx = np.argsort(np.argsort(sim_matrix, axis=1), axis=1)
            sim_matrix_idx = np.argsort(srtd_idx + sim_matrix, axis=0)
        else:
            sim_matrix_idx = np.argsort(sim_matrix, axis=0)

        # Translate the words and for each word return the `topn` similar words
        translated_word = OrderedDict()
        for idx, word in enumerate(source_words):
            translated_target_word = []
            # Search the most `topn` similar words
            for j in range(topn):
                map_space_id = sim_matrix_idx[j, source_space.word2index[word]]
                translated_target_word.append(target_space.index2word[map_space_id])
            translated_word[word] = translated_target_word
        return translated_word


class BackMappingTranslationMatrix(utils.SaveLoad):
    """Realize the BackMapping translation matrix which maps the source model's document vector
    to the target model's document vector (old model).

    BackMapping translation matrix is used to learn a mapping for two document vector spaces which we
    specify as source document vector and target document vector. The target document vectors are trained
    on a superset corpus of source document vectors; we can incrementally increase the vector in
    the old model through the BackMapping translation matrix.

    For details on use, see the tutorial notebook [3]_.

    Examples
    --------
    .. sourcecode:: pycon

        >>> from gensim.test.utils import datapath
        >>> from gensim.test.test_translation_matrix import read_sentiment_docs
        >>> from gensim.models import Doc2Vec, BackMappingTranslationMatrix
        >>>
        >>> data = read_sentiment_docs(datapath("alldata-id-10.txt"))[:5]
        >>> src_model = Doc2Vec.load(datapath("small_tag_doc_5_iter50"))
        >>> dst_model = Doc2Vec.load(datapath("large_tag_doc_10_iter50"))
        >>>
        >>> model_trans = BackMappingTranslationMatrix(src_model, dst_model)
        >>> trans_matrix = model_trans.train(data)
        >>>
        >>> result = model_trans.infer_vector(dst_model.dv[data[3].tags])

    """
    def __init__(self, source_lang_vec, target_lang_vec, tagged_docs=None, random_state=None):
        """

        Parameters
        ----------
        source_lang_vec : :class:`~gensim.models.doc2vec.Doc2Vec`
            Source Doc2Vec model.
        target_lang_vec : :class:`~gensim.models.doc2vec.Doc2Vec`
            Target Doc2Vec model.
        tagged_docs : list of :class:`~gensim.models.doc2vec.TaggedDocument`, optional.
            Documents that will be used for training, both the source language document vector and
            target language document vector trained on those tagged documents.
        random_state : {None, int, array_like}, optional
            Seed for random state.

        """
        self.tagged_docs = tagged_docs
        self.source_lang_vec = source_lang_vec
        self.target_lang_vec = target_lang_vec

        self.random_state = utils.get_random_state(random_state)
        self.translation_matrix = None

        if tagged_docs is not None:
            self.train(tagged_docs)

    def train(self, tagged_docs):
        """Build the translation matrix to map from the source model's vectors to target model's vectors

        Parameters
        ----------
        tagged_docs : list of :class:`~gensim.models.doc2vec.TaggedDocument`, Documents
            that will be used for training, both the source language document vector and
            target language document vector trained on those tagged documents.

        Returns
        -------
        numpy.ndarray
            Translation matrix that maps from the source model's vectors to target model's vectors.

        """
        m1 = [self.source_lang_vec.dv[item.tags].flatten() for item in tagged_docs]
        m2 = [self.target_lang_vec.dv[item.tags].flatten() for item in tagged_docs]

        self.translation_matrix = np.linalg.lstsq(m2, m1, -1)[0]
        return self.translation_matrix

    def infer_vector(self, target_doc_vec):
        """Translate the target model's document vector to the source model's document vector

        Parameters
        ----------
        target_doc_vec : numpy.ndarray
            Document vector from the target document, whose document are not in the source model.

        Returns
        -------
        numpy.ndarray
            Vector `target_doc_vec` in the source model.

        """
        return np.dot(target_doc_vec, self.translation_matrix)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""This module implements functionality related to the `Okapi Best Matching
<https://en.wikipedia.org/wiki/Okapi_BM25>`_ class of bag-of-words vector space models.

Robertson and Zaragoza [1]_ describe the original algorithm and its modifications.

.. [1] Robertson S., Zaragoza H. (2015). `The Probabilistic Relevance Framework: BM25 and
   Beyond, <http://www.staff.city.ac.uk/~sb317/papers/foundations_bm25_review.pdf>`_.

"""

from abc import ABCMeta, abstractmethod
from collections import defaultdict
import logging
import math

from gensim import interfaces, utils
import numpy as np


logger = logging.getLogger(__name__)


class BM25ABC(interfaces.TransformationABC, metaclass=ABCMeta):
    """Objects of this abstract class realize the transformation between word-document co-occurrence
    matrix (int) into a BM25 matrix (positive floats). Concrete subclasses of this abstract class
    implement different BM25 scoring functions.

    """
    def __init__(self, corpus=None, dictionary=None):
        r"""Pre-compute the average length of a document and inverse term document frequencies,
        which will be used to weight term frequencies for the documents.

        Parameters
        ----------
        corpus : iterable of iterable of (int, int) or None, optional
            An input corpus, which will be used to compute the average length of a document and
            inverse term document frequencies. If None, then `dictionary` will be used to compute
            the statistics. If both `corpus` and `dictionary` are None, the statistics will be left
            unintialized. Default is None.
        dictionary : :class:`~gensim.corpora.Dictionary`
            An input dictionary, which will be used to compute the average length of a document and
            inverse term document frequencies.  If None, then `corpus` will be used to compute the
            statistics. If both `corpus` and `dictionary` are None, the statistics will be left
            unintialized. Default is None.

        Attributes
        ----------
        avgdl : float
            The average length of a document.
        idfs : dict of (int, float)
            A mapping from term ids to inverse term document frequencies.

        """
        self.avgdl, self.idfs = None, None
        if dictionary:
            if corpus:
                logger.warning("constructor received both corpus and dictionary; ignoring the corpus")
            num_tokens = sum(dictionary.cfs.values())
            self.avgdl = num_tokens / dictionary.num_docs
            self.idfs = self.precompute_idfs(dictionary.dfs, dictionary.num_docs)
        elif corpus:
            dfs = defaultdict(lambda: 0)
            num_tokens = 0
            num_docs = 0
            for bow in corpus:
                num_tokens += len(bow)
                for term_id in set(term_id for term_id, _ in bow):
                    dfs[term_id] += 1
                num_docs += 1
            self.avgdl = num_tokens / num_docs
            self.idfs = self.precompute_idfs(dfs, num_docs)
        else:
            pass

    @abstractmethod
    def precompute_idfs(self, dfs, num_docs):
        """Precompute inverse term document frequencies, which will be used to weight term frequencies
        for the documents.

        Parameters
        ----------
        dfs : dict of (int, int)
            A mapping from term ids to term document frequencies.
        num_docs : int
            The total number of documents in the training corpus.

        Returns
        -------
        idfs : dict of (int, float)
            A mapping from term ids to inverse term document frequencies.

        """
        pass

    @abstractmethod
    def get_term_weights(self, num_tokens, term_frequencies, idfs):
        """Compute vector space weights for a set of terms in a document.

        Parameters
        ----------
        num_tokens : int
            The number of tokens in the document.
        term_frequencies : ndarray
            1D array of term frequencies.
        idfs : ndarray
            1D array of inverse term document frequencies.

        Returns
        -------
        term_weights : ndarray
            1D array of vector space weights.

        """
        pass

    def __getitem__(self, bow):
        is_corpus, bow = utils.is_corpus(bow)
        if is_corpus:
            return self._apply(bow)

        num_tokens = sum(freq for term_id, freq in bow)

        term_ids, term_frequencies, idfs = [], [], []
        for term_id, term_frequency in bow:
            term_ids.append(term_id)
            term_frequencies.append(term_frequency)
            idfs.append(self.idfs.get(term_id) or 0.0)
        term_frequencies, idfs = np.array(term_frequencies), np.array(idfs)

        term_weights = self.get_term_weights(num_tokens, term_frequencies, idfs)

        vector = [
            (term_id, float(weight))
            for term_id, weight
            in zip(term_ids, term_weights)
        ]
        return vector


class OkapiBM25Model(BM25ABC):
    """The original Okapi BM25 scoring function of Robertson et al. [2]_.

    Examples
    --------
    .. sourcecode:: pycon

        >>> from gensim.corpora import Dictionary
        >>> from gensim.models import OkapiBM25Model
        >>> from gensim.test.utils import common_texts
        >>>
        >>> dictionary = Dictionary(common_texts)  # fit dictionary
        >>> model = OkapiBM25Model(dictionary=dictionary)  # fit model
        >>>
        >>> corpus = [dictionary.doc2bow(line) for line in common_texts]  # convert corpus to BoW format
        >>> vector = model[corpus[0]]  # apply model to the first corpus document

    References
    ----------
    .. [2] Robertson S. E., Walker S., Jones S., Hancock-Beaulieu M. M., Gatford M. (1995).
       `Okapi at TREC-3 <http://research.microsoft.com/pubs/67649/okapi_trec3.pdf>`_.
       *NIST Special Publication 500-226*.

    """
    def __init__(self, corpus=None, dictionary=None, k1=1.5, b=0.75, epsilon=0.25):
        r"""Pre-compute the average length of a document and inverse term document frequencies,
        which will be used to weight term frequencies for the documents.

        Parameters
        ----------
        corpus : iterable of iterable of (int, int) or None, optional
            An input corpus, which will be used to compute the average length of a document and
            inverse term document frequencies. If None, then `dictionary` will be used to compute
            the statistics. If both `corpus` and `dictionary` are None, the statistics will be left
            unintialized. Default is None.
        dictionary : :class:`~gensim.corpora.Dictionary`
            An input dictionary, which will be used to compute the average length of a document and
            inverse term document frequencies.  If None, then `corpus` will be used to compute the
            statistics. If both `corpus` and `dictionary` are None, the statistics will be left
            unintialized. Default is None.
        k1 : float
            A positive tuning parameter that determines the impact of the term frequency on its BM25
            weight. Singhal [5]_ suggests to set `k1` between 1.0 and 2.0. Default is 1.5.
        b : float
            A tuning parameter between 0.0 and 1.0 that determines the document length
            normalization: 1.0 corresponds to full document normalization, while 0.0 corresponds to
            no length normalization. Singhal [5]_ suggests to set `b` to 0.75, which is the default.
        epsilon : float
            A positive tuning parameter that lower-bounds an inverse document frequency.
            Defaults to 0.25.

        Attributes
        ----------
        k1 : float
            A positive tuning parameter that determines the impact of the term frequency on its BM25
            weight. Singhal [3]_ suggests to set `k1` between 1.0 and 2.0. Default is 1.5.
        b : float
            A tuning parameter between 0.0 and 1.0 that determines the document length
            normalization: 1.0 corresponds to full document normalization, while 0.0 corresponds to
            no length normalization. Singhal [3]_ suggests to set `b` to 0.75, which is the default.
        epsilon : float
            A positive tuning parameter that lower-bounds an inverse document frequency.
            Defaults to 0.25.

        References
        ----------
        .. [3] Singhal, A. (2001). `Modern information retrieval: A brief overview
           <http://singhal.info/ieee2001.pdf>`_. *IEEE Data Eng. Bull.*, 24(4), 35–43.

        """
        self.k1, self.b, self.epsilon = k1, b, epsilon
        super().__init__(corpus, dictionary)

    def precompute_idfs(self, dfs, num_docs):
        idf_sum = 0
        idfs = dict()
        negative_idfs = []
        for term_id, freq in dfs.items():
            idf = math.log(num_docs - freq + 0.5) - math.log(freq + 0.5)
            idfs[term_id] = idf
            idf_sum += idf
            if idf < 0:
                negative_idfs.append(term_id)
        average_idf = idf_sum / len(idfs)

        eps = self.epsilon * average_idf
        for term_id in negative_idfs:
            idfs[term_id] = eps

        return idfs

    def get_term_weights(self, num_tokens, term_frequencies, idfs):
        term_weights = idfs * (term_frequencies * (self.k1 + 1)
                              / (term_frequencies + self.k1 * (1 - self.b + self.b
                                                              * num_tokens / self.avgdl)))
        return term_weights


class LuceneBM25Model(BM25ABC):
    """The scoring function of Apache Lucene 8 [4]_.

    Examples
    --------
    .. sourcecode:: pycon

        >>> from gensim.corpora import Dictionary
        >>> from gensim.models import LuceneBM25Model
        >>> from gensim.test.utils import common_texts
        >>>
        >>> dictionary = Dictionary(common_texts)  # fit dictionary
        >>> corpus = [dictionary.doc2bow(line) for line in common_texts]  # convert corpus to BoW format
        >>>
        >>> model = LuceneBM25Model(dictionary=dictionary)  # fit model
        >>> vector = model[corpus[0]]  # apply model to the first corpus document

    References
    ----------
    .. [4] Kamphuis, C., de Vries, A. P., Boytsov, L., Lin, J. (2020). Which
       BM25 Do You Mean? `A Large-Scale Reproducibility Study of Scoring Variants
       <https://doi.org/10.1007/978-3-030-45442-5_4>`_. In: Advances in Information Retrieval.
       28–34.

    """
    def __init__(self, corpus=None, dictionary=None, k1=1.5, b=0.75):
        r"""Pre-compute the average length of a document and inverse term document frequencies,
        which will be used to weight term frequencies for the documents.

        Parameters
        ----------
        corpus : iterable of iterable of (int, int) or None, optional
            An input corpus, which will be used to compute the average length of a document and
            inverse term document frequencies. If None, then `dictionary` will be used to compute
            the statistics. If both `corpus` and `dictionary` are None, the statistics will be left
            unintialized. Default is None.
        dictionary : :class:`~gensim.corpora.Dictionary`
            An input dictionary, which will be used to compute the average length of a document and
            inverse term document frequencies.  If None, then `corpus` will be used to compute the
            statistics. If both `corpus` and `dictionary` are None, the statistics will be left
            unintialized. Default is None.
        k1 : float
            A positive tuning parameter that determines the impact of the term frequency on its BM25
            weight. Singhal [5]_ suggests to set `k1` between 1.0 and 2.0. Default is 1.5.
        b : float
            A tuning parameter between 0.0 and 1.0 that determines the document length
            normalization: 1.0 corresponds to full document normalization, while 0.0 corresponds to
            no length normalization. Singhal [5]_ suggests to set `b` to 0.75, which is the default.

        Attributes
        ----------
        k1 : float
            A positive tuning parameter that determines the impact of the term frequency on its BM25
            weight. Singhal [3]_ suggests to set `k1` between 1.0 and 2.0. Default is 1.5.
        b : float
            A tuning parameter between 0.0 and 1.0 that determines the document length
            normalization: 1.0 corresponds to full document normalization, while 0.0 corresponds to
            no length normalization. Singhal [3]_ suggests to set `b` to 0.75, which is the default.

        """
        self.k1, self.b = k1, b
        super().__init__(corpus, dictionary)

    def precompute_idfs(self, dfs, num_docs):
        idfs = dict()
        for term_id, freq in dfs.items():
            idf = math.log(num_docs + 1.0) - math.log(freq + 0.5)
            idfs[term_id] = idf
        return idfs

    def get_term_weights(self, num_tokens, term_frequencies, idfs):
        term_weights = idfs * (term_frequencies
                              / (term_frequencies + self.k1 * (1 - self.b + self.b
                                                              * num_tokens / self.avgdl)))
        return term_weights


class AtireBM25Model(BM25ABC):
    """The scoring function of Trotman et al. [5]_.

    Examples
    --------
    .. sourcecode:: pycon

        >>> from gensim.corpora import Dictionary
        >>> from gensim.models import AtireBM25Model
        >>> from gensim.test.utils import common_texts
        >>>
        >>> dictionary = Dictionary(common_texts)  # fit dictionary
        >>> corpus = [dictionary.doc2bow(line) for line in common_texts]  # convert corpus to BoW format
        >>>
        >>> model = AtireBM25Model(dictionary=dictionary)  # fit model
        >>> vector = model[corpus[0]]  # apply model to the first corpus document

    References
    ----------
    .. [5] Trotman, A., Jia X., Crane M., `Towards an Efficient and Effective Search Engine
       <http://www.cs.otago.ac.nz/homepages/andrew/involvement/2012-SIGIR-OSIR.pdf#page=45>`_,
       In: SIGIR 2012 Workshop on Open Source Information Retrieval. 40–47.

    """
    def __init__(self, corpus=None, dictionary=None, k1=1.5, b=0.75):
        r"""Pre-compute the average length of a document and inverse term document frequencies,
        which will be used to weight term frequencies for the documents.

        Parameters
        ----------
        corpus : iterable of iterable of (int, int) or None, optional
            An input corpus, which will be used to compute the average length of a document and
            inverse term document frequencies. If None, then `dictionary` will be used to compute
            the statistics. If both `corpus` and `dictionary` are None, the statistics will be left
            unintialized. Default is None.
        dictionary : :class:`~gensim.corpora.Dictionary`
            An input dictionary, which will be used to compute the average length of a document and
            inverse term document frequencies.  If None, then `corpus` will be used to compute the
            statistics. If both `corpus` and `dictionary` are None, the statistics will be left
            unintialized. Default is None.
        k1 : float
            A positive tuning parameter that determines the impact of the term frequency on its BM25
            weight. Singhal [5]_ suggests to set `k1` between 1.0 and 2.0. Default is 1.5.
        b : float
            A tuning parameter between 0.0 and 1.0 that determines the document length
            normalization: 1.0 corresponds to full document normalization, while 0.0 corresponds to
            no length normalization. Singhal [5]_ suggests to set `b` to 0.75, which is the default.

        Attributes
        ----------
        k1 : float
            A positive tuning parameter that determines the impact of the term frequency on its BM25
            weight. Singhal [3]_ suggests to set `k1` between 1.0 and 2.0. Default is 1.5.
        b : float
            A tuning parameter between 0.0 and 1.0 that determines the document length
            normalization: 1.0 corresponds to full document normalization, while 0.0 corresponds to
            no length normalization. Singhal [3]_ suggests to set `b` to 0.75, which is the default.

        """
        self.k1, self.b = k1, b
        super().__init__(corpus, dictionary)

    def precompute_idfs(self, dfs, num_docs):
        idfs = dict()
        for term_id, freq in dfs.items():
            idf = math.log(num_docs) - math.log(freq)
            idfs[term_id] = idf
        return idfs

    def get_term_weights(self, num_tokens, term_frequencies, idfs):
        term_weights = idfs * (term_frequencies * (self.k1 + 1)
                              / (term_frequencies + self.k1 * (1 - self.b + self.b
                                                              * num_tokens / self.avgdl)))
        return term_weights



# -*- coding: utf-8 -*-
"""
Created on Thu Oct 27 11:04:27 2022

@author: 20200016
"""

import math
from collections import Counter
import warnings
import pickle
import itertools
import numpy as np
from scipy.sparse.linalg import svds
from scipy.sparse import dok_matrix
from pyfume import Clustering
import gensim.corpora as corpora
from gensim.models.coherencemodel import CoherenceModel
from gensim.models import Word2Vec


class FlsaModel():
    """
    Class to initialize and train fuzzy topic models with methods similar
    to Gensim's LdaModel'

    Parameters
    ----------
            corpus : The input corpus.
                either: list of list of str.
                or: list of list of tuples (int, int) (bow).

            num_topics: int
                The number of topics to be trained.

            algorithm: str ['flsa', 'flsa-w', 'flsa-e']
                The algorithm to train.

            id2word: gensim.corpora.dictionary.Dictionary
                Object to map id's to words
                (only used when the corpus is passed into the object as a bow).

            word_weighting: str ['normal', 'idf', 'probidf', 'entropy']
                Global term weighting mechanism.

            cluster_method: str ['fcm', 'gk', 'fst-pso']
                Fuzzy clustering method.

            svd_factors: int
                The number of singular values to use.
    """
    def __init__(
            self,
            corpus,
            num_topics,
            algorithm,
            num_words=20,
            word_weighting='normal',
            cluster_method='fcm',
            svd_factors=2,
            id2word=None,
            min_count=None,
            window=None,
            vector_size=None,
            workers=None,
            ):
        self.corpus = self._set_corpus(corpus, id2word)
        self.num_topics = num_topics
        self.algorithm = algorithm
        self.num_topics = num_topics
        self.num_words = num_words
        self.word_weighting = word_weighting
        self.cluster_method = cluster_method
        self.svd_factors = svd_factors
        self.min_count = min_count
        self.window = window
        self.vector_size = vector_size
        self.workers = workers
        self._check_variables()
        self._vocabulary, self._vocabulary_size = self._create_vocabulary(self.corpus)
        self._word_to_index, self._index_to_word = self._create_index_dicts(self._vocabulary)
        self._sum_words = self._create_sum_words(self.corpus)
        self._prob_word_i = None
        self._prob_document_j = None
        self._prob_topic_k = None
        self._prob_word_given_topic = None
        self._prob_word_given_document = None
        self.coherence_score = None
        self.diversity_score = None
        self.pwgt, self.ptgd = self._get_matrices()

    def _set_corpus(
            self,
            corpus,
            id2word,
            ):
        """
        Method that sets the corpus to FuzzyTM's required input format.
        If a list of list of str is passed into the method for corpus, then
        it returns the same corpus. If a bow (list of list of tuples) is passed
        into the class, it transforms this into a list of list of str.

        Parameters
        ----------
            corpus : either: list of list of str (tokens). or: list of list of tuples (int, int).
                The input corpus.
            id2word: gensim.corpora.dictionary.Dictionary
                Object to map id's to words
                (only used when the corpus is passed into the object as a bow)

        Returns
        -------
            list of list of str
                The corpus in FuzzyTM's required input format.
        """
        if self._check_bow(corpus):
            if not isinstance(id2word, corpora.dictionary.Dictionary):
                raise ValueError("Please pass 'id2word' when using a bow for 'corpus'.")
            return self._convert_bow(corpus, id2word)
        return corpus

    @staticmethod
    def _check_bow(
            corpus,
            ):
        """
        Method to check if the input format has the bow format.

        Parameters
        ----------
            corpus : either: list of list of str (tokens). or: list of list of tuples (int, int).
                The input corpus.

        Returns
        -------
            bool
                True if bow format
        """
        if not isinstance(corpus, list):
            return False
        for doc in corpus:
            if not isinstance(doc, list):
                return False
            for tup in doc:
                if not isinstance(tup, tuple):
                    return False
                if not isinstance(tup[0], int) or not isinstance(tup[1], int):
                    return False
        return True

    @staticmethod
    def _convert_bow(
            corpus,
            id2word,
            ):
        """
        Method to convert the bow format into a list of list of str.

        Parameters
        ----------
            corpus : The input corpus.
                either: list of list of str (tokens).
                or: list of list of tuples (int, int).

            id2word: gensim.corpora.dictionary.Dictionary
                Object to map id's to words

        Returns
        -------
            list of list of str
                The corpus in FuzzyTM's required input format.
        """
        id2token = {v: k for k, v in id2word.token2id.items()}
        data_list = []
        for doc in corpus:
            doc_list = []
            for tup in doc:
                for _ in itertools.repeat(None, tup[1]):
                    doc_list.append(id2token[tup[0]])
            data_list.append(doc_list)
        return data_list

    def _check_variables(self):
        """
        Check whether the input data has the right format.

        Correct format: list of list of str (tokens)
        The function raises an error if the format is incorrect.
        """
        for i, doc in enumerate(self.corpus):
            if not isinstance(doc, list):
                raise TypeError("corpus variable at index ",
                                str(i),
                                " is not a list")
            if not len(doc) > 0:
                raise ValueError(
                    "The corpus has an empty list at index ",
                    str(i),
                    " and should contain at least one str value")
            for j, word in enumerate(doc):
                if not isinstance(word, str):
                    raise TypeError(f"Word {j} of document {i} is not a str")
        if not isinstance(self.num_topics, int) or self.num_topics < 1:
            raise ValueError("Please use a positive int for num_topics")
        if not isinstance(self.num_words, int) or self.num_words < 1:
            raise ValueError("Please use a positive int for num_words")
        if self.algorithm in [
                "flsa",
                "flsa-w",
                ] and self.word_weighting not in [
                "entropy",
                "idf",
                "normal",
                "probidf",
                ]:
            warning = ["Invalid word weighting method",
                       "Please choose between:",
                       "'entropy', 'idf', 'normal' and'probidf'",
                       ]
            raise ValueError(' '.join(warning))
        if self.cluster_method not in [
                "fcm",
                "fst-pso",
                "gk",
                ]:
            raise ValueError(
                "Invalid 'cluster_method. Please choose: 'fcm', 'fst-pso' or 'gk'")
        if not isinstance(self.svd_factors, int) and self.svd_factors > 0:
            raise ValueError("Please use a positive int for svd_factors")
        if self.algorithm not in [
                'flsa',
                'flsa-w',
                'flsa-e',
                ]:
            raise ValueError('Please select a correct "algoritm"')

    @staticmethod
    def _create_vocabulary(corpus):
        """
        Create the vocabulary from 'corpus'.

        Parameters
        ----------
             corpus : list of lists of str
                 The input file used to initialize the model.

        Returns
        -------
            set of str
                All the vocabulary words.
        """
        vocabulary = set(el for lis in corpus for el in lis)
        return vocabulary, len(vocabulary)

    @staticmethod
    def _create_index_dicts(vocabulary):
        """
        Create the dictionaries with mappings between words and indices.

        Parameters
        ----------
            vocabulary : set of str
                All the words in the corpus.

        Returns
        -------
            dict of {str : int}
                Dictionary that maps a vocabulary word to and index number.
            dict of {int : str}
                Dictionary that maps an index number to each vocabulary word.
        """
        if not isinstance(vocabulary, set):
            raise ValueError("Please use a 'set' type for 'vocabulary'.")
        word_to_index = dict()
        index_to_word = dict()
        for i, word in enumerate(vocabulary):
            word_to_index[word] = i
            index_to_word[i] = word
        return word_to_index, index_to_word

    @staticmethod
    def _create_sum_words(corpus):
        """
        Creates a Counter object that stores the count of each word in the corpus (corpus).

        Parameters
        ----------
            corpus : list of lists of str
                The input file used to initialize the model.

        Returns
        -------
            collections.Counter {str : int}
                The count of each word in the corpus.
        """
        sum_words = Counter()
        for document in corpus:
            sum_words.update(Counter(document))
        return sum_words

    @staticmethod
    def _create_sparse_local_term_weights(
            corpus,
            vocabulary_size,
            word_to_index,
            ):
        """
        Creates a sparse matrix showing the frequency of each words in documents.

        (See: https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.dok_matrix.html)
        Axes:
            rows: documents (size: number of documents in corpus)
            columns: words (size: vocabulary length)

        Parameters
        ----------
            corpus : list of lists of str
                The input file used to initialize the model.
            vocabulary_size : int
                Number of unique words in the corpus.
            word_to_index: dict {str : int}
                Maps each unique vocabulary word to a unique index number.

        Returns
        -------
            scipy.sparse.dok_matrix
                sparse matrix representation of the local term weights.
        """
        sparse_local_term_weights = dok_matrix(
            (len(corpus),
             vocabulary_size),
            dtype=np.float32,
            )
        for document_index, document in enumerate(corpus):
            document_counter = Counter(document)
            for word in document_counter.keys():
                sparse_local_term_weights[
                    document_index, word_to_index[word],
                    ] = document_counter[word]
        return sparse_local_term_weights

    def _create_sparse_global_term_weights(
            self,
            corpus,
            word_weighting,
            vocabulary_size=None,
            sparse_local_term_weights=None,
            index_to_word=None,
            word_to_index=None,
            sum_words=None,
            ):
        """
        Apply a word_weighting method on the sparse_local_term_weights
        to create sparse_global_term_weights.
        (See: https://link.springer.com/article/10.1007/s40815-017-0327-9)

        Parameters
        ----------
            corpus : list of lists of str
                The input file used to initialize the model.
            word_weighting : str
                Indicates the method used for word_weighting. Choose from:
                  - entropy
                  - normal
                  - idf
                  - probidf
             vocabulary_size : int
                Number of unique words in the corpus.
             sparse_local_term_weights : scipy.sparse.dok_matrix
                 A sparse matrix showing the frequency of each words in documents.
             word_to_index : dict {str : int}
                Maps each unique vocabulary word to a unique index number.
             index_to_word : dict {int : str}
                 Maps each unique index number to a unique vocabulary word.
             sum_words : collections.Counter {str : int}
                 The count of each word in the corpus.

        Returns
        -------
            scipy.sparse.dok_matrix
                sparse matrix representation of the global term weights.
        """
        num_documents = len(corpus)
        if word_weighting in ['entropy', 'normal']:
            if sparse_local_term_weights is None:
                raise ValueError("Please feed the algorithm 'sparse_local_term_weights'")
        if word_weighting in ['entropy']:
            if index_to_word is None:
                raise ValueError("Please feed the algorithm 'index_to_word'")
            if sum_words is None:
                raise ValueError("Please feed the algorithm 'sum_words'")
        if word_weighting in ['entropy', 'idf', 'probidf']:
            if vocabulary_size is None:
                raise ValueError("Please feed the algorithm 'vocabulary_size'")
        if word_weighting in ['idf', 'probidf']:
            if word_to_index is None:
                raise ValueError("Please feed the algorithm 'word_to_index'")
        if word_weighting == 'entropy':
            global_term_weights = self._calculate_entropy(
                num_documents,
                vocabulary_size,
                sparse_local_term_weights,
                index_to_word, sum_words,
                )
        elif word_weighting == 'idf':
            global_term_weights = self._calculate_idf(
                num_documents,
                vocabulary_size,
                corpus,
                word_to_index,
                )
        elif word_weighting == 'normal':
            global_term_weights = self._calculate_normal(sparse_local_term_weights)
        elif word_weighting == 'probidf':
            global_term_weights = self._calculate_probidf(
                num_documents,
                vocabulary_size,
                corpus,
                word_to_index,
                )
        else:
            raise ValueError('Invalid word weighting method')
        return sparse_local_term_weights.multiply(global_term_weights).tocsc()

    def _calculate_entropy(
            self,
            num_documents,
            vocabulary_size,
            sparse_local_term_weights,
            index_to_word,
            sum_words,
            ):
        """
        Use the entropy word weighting method.

        (See: https://link.springer.com/article/10.1007/s40815-017-0327-9)

        Parameters
        ----------
            num_documents : int
                The number of documents in the corpus.
            vocabulary_size : int
                Number of unique words in the corpus.
            sparse_local_term_weights : scipy.sparse.dok_matrix
                 A sparse matrix showing the frequency of each words in documents.
            index_to_word : dict {int : str}
                 Maps each unique index number to a unique vocabulary word.
            sum_words : collections.Counter {str : int}
                 The count of each word in the corpus.

        Returns
        -------
            numpy.array : float
        """
        p_log_p_ij = self._create_p_log_p_ij(
            num_documents,
            vocabulary_size,
            sparse_local_term_weights,
            index_to_word,
            sum_words,
            )
        summed_p_log_p = p_log_p_ij.sum(0).tolist()[0]
        return np.array([1 + summed_p_log_p_i / np.log2(num_documents) for summed_p_log_p_i in summed_p_log_p])

    def _calculate_idf(
            self,
            num_documents,
            vocabulary_size,
            corpus,
            word_to_index,
            ):
        """
        Use the idf word weightingg method.

        (See: https://link.springer.com/article/10.1007/s40815-017-0327-9)

        Parameters
        ----------
            num_documents : int
                The number of documents in the corpus.
            vocabulary_size : int
                Number of unique words in the corpus.
            corpus : list of lists of str
                The input file used to initialize the model.
            word_to_index: dict {str : int}
                Maps each unique vocabulary word to a unique index number.

        Returns
        -------
            numpy.array : float
        """
        binary_sparse_dtm = self._create_sparse_binary_dtm(
            num_documents,
            vocabulary_size,
            corpus,
            word_to_index,
            )
        summed_words = binary_sparse_dtm.sum(0).tolist()[0]
        return np.array([np.log2(num_documents / word_count) for word_count in summed_words])

    @staticmethod
    def _calculate_normal(
            sparse_local_term_weights,
            ):
        """
        Use the normal word weightingg method.

        (See: https://link.springer.com/article/10.1007/s40815-017-0327-9)

        Parameters
        ----------
             sparse_local_term_weights : scipy.sparse.dok_matrix
                 A sparse matrix showing the frequency of each words in documents.

        Returns
        -------
            numpy.array : float
        """
        squared_dtm = sparse_local_term_weights.multiply(sparse_local_term_weights)
        summed_words = squared_dtm.sum(0).tolist()[0]
        return np.array([1 / (math.sqrt(word_count)) for word_count in summed_words])

    def _calculate_probidf(
            self,
            num_documents,
            vocabulary_size,
            corpus,
            word_to_index,
            ):
        """
        Use the probidf word weightingg method.

        (See: https://link.springer.com/article/10.1007/s40815-017-0327-9)

        Parameters
        ----------
            num_documents : int
                The number of documents in the corpus.
            vocabulary_size : int
                Number of unique words in the corpus.
            corpus : list of lists of str
                The input file used to initialize the model.
            word_to_index: dict {str : int}
                Maps each unique vocabulary word to a unique index number.

        Returns
        -------
            numpy.array : float
        """
        binary_sparse_dtm = self._create_sparse_binary_dtm(
            num_documents,
            vocabulary_size,
            corpus,
            word_to_index,
            )
        summed_binary_words_list = binary_sparse_dtm.sum(0).tolist()[0]

        return np.array([np.log2((num_documents - binary_word_count) / binary_word_count)
                         for binary_word_count in summed_binary_words_list])

    @staticmethod
    def _create_p_log_p_ij(
            num_documents,
            vocabulary_size,
            sparse_local_term_weights,
            index_to_word,
            sum_words,
            ):
        """
        Create probability of word i in document j, multiplied by its base-2 logarithm.

        (See: https://link.springer.com/article/10.1007/s40815-017-0327-9)

        Parameters
        ----------
            num_documents : int
                The number of documents in the corpus.
            vocabulary_size : int
                Number of unique words in the corpus.
            sparse_local_term_weights : scipy.sparse.dok_matrix
                A sparse matrix showing the frequency of each words in documents.
            index_to_word : dict {int : str}
                Maps each unique index number to a unique vocabulary word.
            sum_words : collections.Counter {str : int}
                The count of each word in the corpus.

        Returns
        -------
             scipy.sparse.dok_matrix
        """
        p_log_p_ij = dok_matrix(
            (num_documents, vocabulary_size), dtype=np.float32,
            )
        for j in range(num_documents):
            row_counts = sparse_local_term_weights.getrow(j).toarray()[0]
            word_index = row_counts.nonzero()[0]
            non_zero_row_counts = row_counts[row_counts != 0]
            for i, count in enumerate(non_zero_row_counts):
                word = index_to_word[word_index[i]]
                prob_ij = count / sum_words[word]
                p_log_p_ij[j, word_index[i]] = prob_ij * np.log2(prob_ij)
        return p_log_p_ij

    @staticmethod
    def _create_sparse_binary_dtm(
            num_documents,
            vocabulary_size,
            corpus,
            word_to_index,
            ):
        """
        Create a binary sparse document-term-matrix (used for idf and probidf).

        (See: https://link.springer.com/article/10.1007/s40815-017-0327-9)

        Parameters
        ----------
            num_documents : int
                The number of documents in the corpus.
            vocabulary_size : int
                Number of unique words in the corpus.
            corpus : list of lists of str
                The input file used to initialize the model.
            word_to_index: dict {str : int}
                Maps each unique vocabulary word to a unique index number.

       Returns
        -------
             scipy.sparse.dok_matrix
        """
        binary_sparse_dtm = dok_matrix(
            (num_documents, vocabulary_size), dtype=np.float32,
            )
        for doc_index, document in enumerate(corpus):
            binary_document_counter = dict.fromkeys(document, 1)
            for word in set(document):
                binary_sparse_dtm[doc_index,
                                  word_to_index[word]] = binary_document_counter[word]
        return binary_sparse_dtm

    @staticmethod
    def _create_projected_data(
            algorithm,
            sparse_weighted_matrix,
            svd_factors,
            ):
        """
        Perform singular decomposition for dimensionality reduction.

        (See: https://web.mit.edu/be.400/www/SVD/Singular_Value_Decomposition.htm)
        For SVD on a sparse matrix, the sparsesvd package is used
        (https://pypi.org/project/sparsesvd/)

        Parameters
        ----------
             algorithm : str
                 Indicator for which algorithm is being trained ('flsa' or 'flsa-w').
             sparse_weighted_matrix : scipy.sparse.dok_matrix
                 Sparse global term matrix.
             svd_factors : int
                 The number of singular values to include.

        Returns
        -------
            numpy.array : float
        """
        svd_u, _, svd_v = svds(
            sparse_weighted_matrix,
            svd_factors,
            )
        if algorithm in ['flsa']:
            return svd_u
        if algorithm in ['flsa-w']:
            return svd_v.T
        raise ValueError('Invalid algorithm selected.',
                         'Only "flsa" ans "flsa-w" are currently supported.')

    @staticmethod
    def _create_partition_matrix(
            data,
            number_of_clusters,
            method='fcm',
            ):
        """
        Perform clustering on the projected data.

        The pyFUME package is used for clustering:
            (https://pyfume.readthedocs.io/en/latest/Clustering.html)

        Parameters
        ----------
             data: numpy.array
                 The output from self._create_projected_data().
             number_of_clusters : int
                 The number of clusters (topics).
             method : str
                 The cluster method, choose from: 'fcm', 'gk', 'fst-pso'.
        Returns
        -------
            numpy.array : float
        """
        clusterer = Clustering.Clusterer(
            nr_clus=number_of_clusters,
            data=data,
            )
        _, partition_matrix, _ = clusterer.cluster(method=method)
        return partition_matrix

    @staticmethod
    def _create_prob_document_j(sparse_matrix):
        """
        Get the probability of document j.

        Parameters
        ----------
            sparse_matrix : scipy.sparse.dok_matrix
                A sparse matrix representation of the global term weights.
        Returns
        -------
            numpy.array : float
                (shape: number of documents x 1)
        """
        # Vector with the length of num_document,
        # each cell represents the sum of all weights of a document
        document_sum = np.array([doc[0] for doc in sparse_matrix.sum(1).tolist()])
        # sum of all the elements in the weighted matrix
        total_sum_d = sum(sparse_matrix.sum(0).tolist()[0])
        return document_sum / total_sum_d  # normalized probability

    @staticmethod
    def _create_prob_word_i(sparse_matrix):
        """
        Get the probability of word i.

        Parameters
        ----------
            sparse_matrix : scipy.sparse.dok_matrix
                A sparse matrix representation of the global term weights.

        Returns
        -------
            numpy.array : float
                (shape: vocabulary_size x 1)
        """
        word_sum = np.array(sparse_matrix.sum(0).tolist())
        # Sum of all the elements in the weighted matrix
        total_sum_w = sum(sparse_matrix.sum(0).tolist()[0])
        return (word_sum / total_sum_w)[0]  # normalized probability

    @staticmethod
    def _create_prob_topic_k(
            prob_topic_given_word_transpose,
            prob_word_i,
            ):
        """
        Get the probability of topic k.

        Parameters
        ----------
            prob_topic_given_word_transpose : numpy.array : float
                The output from self._create_partition_matrix().
            prob_word_i : numpy.array : float
                The output from self._create_prob_word_i().

        Returns
        -------
            numpy.array : float
                (shape: 1 x number of topics)
        """
        return np.matmul(prob_topic_given_word_transpose.T, prob_word_i)

    @staticmethod
    def _check_passed_variables(
            algorithm,
            prob_topic_given_document_transpose,
            prob_topic_given_word_transpose,
            local_term_weights,
            global_term_weights,
            ):
        """
        Check whether the algorithms are being fed the right attributes.
        """
        if algorithm in ['flsa']:
            if prob_topic_given_document_transpose is None:
                raise ValueError("Please feed the method",
                                 "'prob_topic_given_document_transpose' to run flsa")
            if global_term_weights is None:
                raise ValueError("Please feed the method 'global_term_weights', to run flsa")
        elif algorithm in ['flsa-w']:
            if prob_topic_given_word_transpose is None:
                raise ValueError("Please feed the method",
                                 "'prob_topic_given_word_transpose' to run flsa-w")
            if global_term_weights is None:
                raise ValueError("Please feed the method 'global_term_weights'",
                                 " to run flsa-w")
        elif algorithm in [
                           'flsa-e',
                           ]:
            if prob_topic_given_word_transpose is None:
                raise ValueError("Please feed the method",
                                 "'prob_topic_given_word_transpose' to run model")
            if local_term_weights is None:
                raise ValueError("Please feed the method 'local_term_weights', to run model")

        else:
            raise ValueError('Your algorithm is currently not supported')

    def _create_probability_matrices(
            self,
            algorithm,
            prob_topic_given_document_transpose=None,
            prob_topic_given_word_transpose=None,
            local_term_weights=None,
            global_term_weights=None,
            ):
        """
        Method that performs matrix multiplications to obtain the output matrices.

        The 'algorithm' parameter is generic and the other ones depend on the selected algorithm.
        The other parameters passed into this method depend on the used algorithm.

        Parameters
        ----------
            algorithm : str
                 Indicator for which algorithm is being trained ('flsa' or 'flsa-w').
            global_term_weights : scipy.sparse.dok_matrix
                The output from self._create_partition_matrix().
            prob_topic_given_document_transpose : numpy.array : float
                The output from self._create_partition_matrix() (flsa)
            prob_topic_given_word_transpose : numpy.array : float
                 (flsa-w)

        Returns
        -------
            numpy.array : float
                The prbability of a word given a topic.
            numpy.array : float
                The prbability of a topic given a document.
        """
        # Check whether the right variable are passed into the method.
        self._check_passed_variables(
            algorithm,
            prob_topic_given_document_transpose,
            prob_topic_given_word_transpose,
            local_term_weights,
            global_term_weights,
            )

        # Calculate the initial probabilities
        if algorithm in [
                'flsa',
                'flsa-w',
                ]:
            self._prob_word_i = self._create_prob_word_i(global_term_weights)
            self._prob_document_j = self._create_prob_document_j(global_term_weights)
            if algorithm in ['flsa-w']:
                self._prob_topic_k = self._create_prob_topic_k(
                    prob_topic_given_word_transpose,
                    self._prob_word_i,
                    )
        elif algorithm in [
                           'flsa-e',
                           ]:
            self._prob_word_i = self._create_prob_word_i(local_term_weights)
            self._prob_document_j = self._create_prob_document_j(local_term_weights)
            self._prob_topic_k = self._create_prob_topic_k(
                prob_topic_given_word_transpose, self._prob_word_i,
                )
        if algorithm in ['flsa']:
            prob_document_and_topic = (prob_topic_given_document_transpose.T * self._prob_document_j).T
            prob_document_given_topic = prob_document_and_topic / prob_document_and_topic.sum(axis=0)
            self._prob_word_given_document = np.asarray(global_term_weights / global_term_weights.sum(1))
            self._prob_word_given_topic = np.matmul(
                self._prob_word_given_document.T,
                prob_document_given_topic,
                )
            prob_topic_given_document = prob_topic_given_document_transpose.T
            return self._prob_word_given_topic, prob_topic_given_document

        elif algorithm in [
                'flsa-w',
                'flsa-e'
                ]:
            prob_word_and_topic = (prob_topic_given_word_transpose.T * self._prob_word_i).T
            self._prob_word_given_topic = prob_word_and_topic / prob_word_and_topic.sum(axis=0)
            if algorithm in ['flsa-w']:
                self._prob_word_given_document = np.asarray(global_term_weights / global_term_weights.sum(1)).T
            elif algorithm in [
                               'flsa-e',
                               ]:
                self._prob_word_given_document = np.asarray(local_term_weights / local_term_weights.sum(1)).T
            prob_document_given_word = ((self._prob_word_given_document * self._prob_document_j).T
                                        / np.array(self._prob_word_i))
            prob_document_given_topic = np.matmul(
                prob_document_given_word,
                self._prob_word_given_topic,
                )
            prob_topic_given_document = ((prob_document_given_topic * self._prob_topic_k).T
                                               / self._prob_document_j)
            return self._prob_word_given_topic, prob_topic_given_document
        raise ValueError('"algorithm" is unknown.')

    @staticmethod
    def _create_dictlist_topn(
            topn,
            prob_word_given_topic,
            index_to_word,
            ):
        """
        Creates a list with dictionaries of word probabilities
        per topic based on the top-n words.

        Parameters
        ----------
             topn : int
                The top-n words to include
                (needs only to be used when 'method=topn').
             prob_word_given_topic : numpy.array : float
                Matrix that gives the probability of a word given a topic.
             index_to_word : dict {int : str}
                Maps each unique index number to a unique vocabulary word.

        Returns
        -------
             list of dicts {int : float}
                Keys: all the indices of words from prob_word_given_topic
                who's weight's are amongst the top percentage.
                Values: the probability associated to a word.
        """
        if not isinstance(topn, int) and topn > 0:
            raise ValueError("Please choose a positive integer for 'topn'")
        top_dictionaries = []
        for topic_index in range(prob_word_given_topic.shape[1]):
            new_dict = dict()
            highest_weight_indices = prob_word_given_topic[:, topic_index].argsort()[-topn:]
            for word_index in highest_weight_indices:
                new_dict[index_to_word[word_index]] = prob_word_given_topic[
                    word_index, topic_index,
                    ]
            top_dictionaries.append(new_dict)
        return top_dictionaries

    @staticmethod
    def _create_dictlist_percentile(
            perc,
            prob_word_given_topic,
            index_to_word,
            ):
        """
        Create a list with dictionaries of word probabilities
        per topic based on the percentile.
         - Keys: all the indices of words from prob_word_given_topic
             who's weight's are amongst the top percentage.
         - Values: the probability associated to a word.

        Parameters
        ----------
             perc : float
                The top percentile words to include
                (needs only to be used when 'method=percentile').
             prob_word_given_topic : numpy.array : float
                Matrix that gives the probability of a word given a topic.
             index_to_word : dict {int : str}
                Maps each unique index number to a unique vocabulary word.

        Returns
        -------
             list of dicts {int : float}
                Keys: all the indices of words from prob_word_given_topic
                    who's weight's are amongst the top percentage.
                Values: the probability associated to a word.
        """
        if not isinstance(perc, float) and 0 <= perc <= 1:
            raise ValueError("Please choose a number between 0 and 1 for 'perc'")
        top_list = []
        for top in range(prob_word_given_topic.shape[1]):
            new_dict = dict()
            count = 0
            i = 0
            weights = np.sort(prob_word_given_topic[:, top])[::-1]
            word_indices = np.argsort(prob_word_given_topic[:, top])[::-1]
            while count < perc:
                new_dict[index_to_word[word_indices[i]]] = weights[i]
                count += weights[i]
                i += 1
            top_list.append(new_dict)
        return top_list

    def show_topics(
            self,
            formatted=True,
            prob_word_given_topic=None,
            num_words=-1,
            index_to_word=None,
            ):
        """
        Get a representation for the topics.

        Parameters
        ----------
            formatted : bool
                Whether the topic representations should be formatted as strings.
                If False, they are returned as 2 tuples of (word, probability).
            prob_word_given_topic : numpy.array : float
                Matrix that gives the probability of a word given a topic.
            num_words : int
                 Indicates how many words per topic should be shown.
            index_to_word : dict {int : str}
                Maps each unique index number to a unique vocabulary word.

        Returns
        -------
            list of tuples (int, str)
                The produced topics.
        """
        if prob_word_given_topic is None:
            prob_word_given_topic = self._prob_word_given_topic

        if num_words < 0:
            num_words = self.num_words
        if index_to_word is None:
            index_to_word = self._index_to_word
        if not isinstance(prob_word_given_topic, np.ndarray):
            raise TypeError("Please feed the algorithm 'prob_word_given_topic' as a np.ndarray")
        if not isinstance(index_to_word, dict):
            raise TypeError("Please feed the algorithm 'index_to_word' as a dict")
        if not isinstance(num_words, int) or num_words <= 0:
            raise TypeError("Please use a positive int for 'num_words'.")
        if prob_word_given_topic.shape[0] < prob_word_given_topic.shape[1]:
            raise ValueError("'prob_word_given_topic' has more columns then rows,",
                             " probably you need to take the transpose.")
        warning = ["It seems like 'prob_word_given_topic' and 'index_to_word",
                   "are not aligned. The number of vocabulary words in",
                   "'prob_word_given_topic' deviate from the ",
                   "number of words in 'index_to_word'."]
        if prob_word_given_topic.shape[0] != len(index_to_word.keys()):
            warnings.warn(' '.join(warning))
        if not isinstance(formatted, bool):
            raise ValueError('Please choose a boolean for "formatted"')
        topic_list = []
        if not formatted:
            for topic_index in range(prob_word_given_topic.shape[1]):
                weight_words = ""
                sorted_highest_weight_indices = prob_word_given_topic[:, topic_index].argsort()[-num_words:][::-1]
                for word_index in sorted_highest_weight_indices:
                    weight_words += (str(round(prob_word_given_topic[word_index, topic_index], 4))
                                     + '*"' + index_to_word[word_index] + '" + ')
                topic_list.append((topic_index, weight_words[:-3]))
            return topic_list
        else:
            for topic_index in range(prob_word_given_topic.shape[1]):
                word_list = []
                sorted_highest_weight_indices = prob_word_given_topic[:, topic_index].argsort()[-num_words:][::-1]
                for word_index in sorted_highest_weight_indices:
                    word_list.append(index_to_word[word_index])
                topic_list.append(word_list)
            return topic_list

    def get_topic_embedding(
            self,
            corpus,
            prob_word_given_topic=None,
            method='topn',
            topn=20,
            perc=0.05,
            ):
        """
        Create a topic embedding for each input document,
        to be used as input to predictive models.

        Parameters
        ----------
            corpus : list of lists of str
                The input file used to initialize the model.
            prob_word_given_topic : numpy.array : float
                Matrix that gives the probability of a word given a topic.
            method : str
                Method to select words to be included in the embedding.
                (choose from 'topn', 'percentile'):
                    - topn: for each topic the top n words with the highest
                        probability are included.
                    - percentile: for each topic all words with highest
                        probabilities are assigned while the cumulative
                        probability is lower than the percentile.
            topn : int
                The top-n words to include
                (needs only to be used when 'method=topn').
            perc: float
                The benchmark percentile until which words need to be added
                (between 0 and 1).

        Returns
        -------
            numpy.array : float
                Array in which each row gives the topic embedding for
                the associated document.
        """
        self._check_variables()
        if prob_word_given_topic is None:
            prob_word_given_topic = self._prob_word_given_topic
        top_dist = []
        if method not in ['topn', 'percentile']:
            raise ValueError(method, "is not a valid option for 'method'.",
                             " Choose either 'topn' or 'percentile'")
        if method == 'topn':
            dictlist = self._create_dictlist_topn(
                topn, prob_word_given_topic, self._index_to_word,
                )
        else:
            dictlist = self._create_dictlist_percentile(
                perc, prob_word_given_topic, self._index_to_word,
                )
        for doc in corpus:
            topic_weights = [0] * prob_word_given_topic.shape[1]
            for word in doc:
                for i in range(prob_word_given_topic.shape[1]):
                    topic_weights[i] += dictlist[i].get(word, 0)
            top_dist.append(topic_weights)
        return np.array(top_dist)

    def get_coherence_score(
            self,
            corpus=None,
            topics=None,
            coherence='c_v',
            ):
        """
        Calculate the coherence score for the generated topic.

        Parameters
        ----------
             corpus : list of lists of str
                The input file used to initialize the model.
             topics : list of lists of str
                 The words per topics,
                 equivalent to self.show_topics(formatted=True).
             coherence : str
                 The type of coherence to be calculated.
                 Choose from: 'u_mass', 'c_v', 'c_uci', 'c_npmi'.

        Returns
        -------
             float
                 The coherence score.
        """
        if corpus is None and topics is None:
            corpus = self.corpus
            topics = self.show_topics(formatted=True)

        id2word = corpora.Dictionary(corpus)
        corpus_bow = [id2word.doc2bow(text) for text in corpus]
        self.coherence_score = CoherenceModel(
            topics=topics,
            texts=corpus,
            corpus=corpus_bow,
            dictionary=id2word,
            coherence=coherence,
            topn=len(topics[0]),
            ).get_coherence()
        return self.coherence_score

    def get_diversity_score(
            self,
            topics=None,
            ):
        """''
        Calculate the diversity score for the generated topic.

        Diversity = number of unique words / number of total words.
        See: https://tinyurl.com/2bs84zd8

        Parameters
        ----------
             topics : list of lists of str
                 The words per topics,
                 equivalent to self.show_topics(formatted=True).

        Returns
        -------
             float
                 The diversity score.
        """
        if topics is None:
            topics = self.show_topics(formatted=True)
        unique_words = set()
        total_words = 0
        for top in topics:
            unique_words.update(top)
            total_words += len(top)
        self.diversity_score = len(unique_words) / total_words
        return self.diversity_score

    def get_interpretability_score(
            self,
            corpus=None,
            topics=None,
            coherence='c_v',
            ):
        """''
        Calculate the interpretability score for the generated topics.

        Interpretability = coherence * diversity.
        (see: https://tinyurl.com/2bs84zd8)

        Parameters
        ----------
             corpus : list of lists of str
                The input file used to initialize the model.
             topics : list of lists of str
                 The words per topics, equivalent to
                 self.show_topics(formatted=True).
             coherence : str
                 The type of coherence to be calculated.
                 Choose from: 'u_mass', 'c_v', 'c_uci', 'c_npmi'.

        Returns
        -------
             float
                 The interpretability score.
        """
        if corpus is None and topics is None:
            corpus = self.corpus
            topics = self.show_topics(formatted=True)
        if self.coherence_score is None:
            self.coherence_score = self.get_coherence_score(
                corpus,
                topics,
                coherence,
                )
        if self.diversity_score is None:
            self.diversity_score = self.get_diversity_score(topics)
        return self.coherence_score * self.diversity_score

    def get_vocabulary(self):
        """
        Returns a set of all the words in the corpus

        Example:
        After initializing an instance of the flsamodel models as 'model'

        corpus = [['this', 'is', 'the', 'first', 'file'],
             ['and', 'this', 'is', 'second', 'file']]

        model.get_vocabulary()

        >>> {'this', 'is', 'the', 'first', 'file', 'and', 'second'}
        """
        return self._vocabulary

    def get_topics(self):
        """
        Get the term-topic matrix.

        Returns
        -------
        numpy.ndarray
            The probability for each word in each topic,
            shape (num_topics, vocabulary_size).
        """
        return self.pwgt

    def get_vocabulary_size(self):
        """
        Returns the number of words in the vocabulary

        Example:
            After initializing an instance of the flsamodel models as 'model'

        corpus = [['this', 'is', 'the', 'first', 'file'],
             ['and', 'this', 'is', 'second', 'file']]

        model.get_vocabulary_size()

        >>> 7
        """
        return self._vocabulary_size

    def get_word_to_index(self):
        """
        Obtain a dictionary that maps each vocabulary word to an index.

        Returns
        -------
        dict of {str : int}
            word to int mapping.
        """
        return self._word_to_index

    def get_index_to_word(self):
        """
        Obtain a dictionary that maps index numbers to vocabulary words.

        Returns
        -------
        dict of {int : str}
            int to word mapping.
        """
        return self._index_to_word

    def get_corpus(self):
        """
        Return the input file.

        Returns
        -------
            list of list of str
                The input file 'corpus'.
        """
        return self.corpus

    def get_prob_word_i(self):
        """
        Return the probabilities per word.

        Returns
        -------
            np.array of float
                The probabilities per word.
        """
        return self._prob_word_i

    def get_prob_document_j(self):
        """
        Return the probabilities per document.

        Returns
        -------
            np.array of float
                The probabilities per document.
        """
        return self._prob_document_j

    def get_prob_topic_k(self):
        """
        Return the probabilities per topic.

        Returns
        -------
            np.array of float
                The probabilities per topic.
        """
        return self._prob_topic_k

    def save(
            self,
            filepath,
            ):
        """''
        Saves the object to the drive, using the pickle library.

        Parameters
        ----------
             filepath : str
                The directory in which the file should be stored,
                either with or without the file name.

        Returns
        -------
             float
                 The interpretability score.
        """
        if not isinstance(filepath, str):
            raise ValueError('Make sure that "filepath" has type "str"')
        if filepath.endswith('.pickle'):
            pickle_out = open(filepath, 'wb')
        elif filepath.endswith('/'):
            pickle_out = open(filepath + 'model.pickle', 'wb')
        else:
            pickle_out = open(filepath + '.pickle', 'wb')
        pickle.dump(self, pickle_out)
        pickle_out.close()

    def load(
            self,
            filepath,
            ):
        """''
        Loads the object from the drive, using the pickle library.

        Parameters
        ----------
             filepath : str
                The directory in which the file should be stored,
                either with or without the file name.

        Returns
        -------
             float
                 The interpretability score.
        """
        if not isinstance(filepath, str):
            raise ValueError('Make sure that "filepath" has type "str"')
        if not filepath.endswith('.pickle'):
            if filepath.endswith('/'):
                filepath += 'model.pickle'
            else:
                filepath += '/model.pickle'
        infile = open(filepath, 'rb')
        self.__dict__ = pickle.load(infile).__dict__
        infile.close()


class Flsa(FlsaModel):
    """
    Class to run the FLSA algorithm (see: https://tinyurl.com/mskjaeuu).

    Parameters
        ----------
            corpus : list of lists of str
                The input file used to initialize the model.
            num_topics : int
                The number of topics that the model should train.
            num_words : int
                 Indicates how many words per topic should be shown.
            word_weighting : str
                Indicates the method used for word_weighting. Choose from:
                  - entropy
                  - normal
                  - idf
                  - probidf
            svd_factors : int
                 The number of singular values to include.
            cluster_method : str
                 The cluster algorithm to be used ('fcm', 'gk', 'fst-pso').
    """
    def __init__(
            self,
            corpus,
            num_topics,
            num_words=10,
            word_weighting='normal',
            svd_factors=2,
            cluster_method='fcm',
            ):
        super().__init__(
            algorithm='flsa',
            corpus=corpus,
            num_topics=num_topics,
            num_words=num_words,
            word_weighting=word_weighting,
            cluster_method=cluster_method,
            svd_factors=svd_factors,
            )

    def _get_matrices(self):
        """
        Method to obtain the matrices after the model has been initialized.

        Returns
        -------
            numpy.array : float
                The prbability of a word given a topic.
            numpy.array : float
                The prbability of a topic given a document.
        """
        sparse_document_term_matrix = self._create_sparse_local_term_weights(
            self.corpus,
            self._vocabulary_size,
            self._word_to_index,
            )
        sparse_global_term_weighting = self._create_sparse_global_term_weights(
            corpus=self.corpus,
            word_weighting=self.word_weighting,
            vocabulary_size=self._vocabulary_size,
            sparse_local_term_weights=sparse_document_term_matrix,
            index_to_word=self._index_to_word,
            word_to_index=self._word_to_index,
            sum_words=self._sum_words,
            )
        projected_data = self._create_projected_data(
            algorithm='flsa',
            sparse_weighted_matrix=sparse_global_term_weighting,
            svd_factors=self.svd_factors,
            )
        partition_matrix = self._create_partition_matrix(
            data=projected_data,
            number_of_clusters=self.num_topics,
            method=self.cluster_method
            )
        return self._create_probability_matrices(
            algorithm='flsa',
            prob_topic_given_document_transpose=partition_matrix,
            global_term_weights=sparse_global_term_weighting,
            )


class FlsaW(FlsaModel):
    """
    Class to train the FLSA-W algorithm.

    See: https://ieeexplore.ieee.org/abstract/document/9660139

    Parameters
        ----------
            corpus : list of lists of str
                The input file used to initialize the model.
            num_topics : int
                The number of topics that the model should train.
            num_words : int
                 Indicates how many words per topic should be shown.
            word_weighting : str
                Indicates the method used for word_weighting. Choose from:
                  - entropy
                  - normal
                  - idf
                  - probidf
            svd_factors : int
                 The number of singular values to include.
            cluster_method : str
                 The cluster algorithm to be used ('fcm', 'gk', 'fst-pso').
    """
    def __init__(
            self,
            corpus,
            num_topics,
            num_words=10,
            word_weighting='normal',
            svd_factors=2,
            cluster_method='fcm',
            ):

        super().__init__(
            algorithm='flsa-w',
            corpus=corpus,
            num_topics=num_topics,
            num_words=num_words,
            word_weighting=word_weighting,
            cluster_method=cluster_method,
            svd_factors=svd_factors,
            )

    def _get_matrices(self):
        """
        Method to obtain the matrices after the model has been initialized.

        Returns
        -------
            numpy.array : float
                The prbability of a word given a topic.
            numpy.array : float
                The prbability of a topic given a document.
        """
        sparse_document_term_matrix = self._create_sparse_local_term_weights(
            self.corpus,
            self._vocabulary_size,
            self._word_to_index,
            )
        sparse_global_term_weighting = self._create_sparse_global_term_weights(
            corpus=self.corpus,
            word_weighting=self.word_weighting,
            vocabulary_size=self._vocabulary_size,
            sparse_local_term_weights=sparse_document_term_matrix,
            index_to_word=self._index_to_word,
            word_to_index=self._word_to_index,
            sum_words=self._sum_words,
            )
        projected_data = self._create_projected_data(
            algorithm='flsa-w',
            sparse_weighted_matrix=sparse_global_term_weighting,
            svd_factors=self.svd_factors,
            )
        partition_matrix = self._create_partition_matrix(
            data=projected_data,
            number_of_clusters=self.num_topics,
            method=self.cluster_method,
            )
        return self._create_probability_matrices(
            algorithm='flsa-w',
            prob_topic_given_word_transpose=partition_matrix,
            global_term_weights=sparse_global_term_weighting,
            )


class FlsaE(FlsaModel):
    """
    Class to train the FLSA-E algorithm. See: https://tinyurl.com/5n8utppk

    Parameters
        ----------
            corpus : list of lists of str
                The input file used to initialize the model.
            num_topics : int
                The number of topics that the model should train.
            num_words : int
                Indicates how many words per topic should be shown.
            cluster_method : str
                The cluster algorithm to be used ('fcm', 'gk', 'fst-pso').
            min_count : int
                Ignores all words with total frequency lower than this.
            window : int
                Maximum distance between the current and predicted word within a sentence.
            vector_size : int
                Dimensionality of the word vectors.
            workers : int
                Use these many worker threads to train the model
                ( = faster training with multicore machines).
    """

    def __init__(
            self,
            corpus,
            num_topics,
            num_words=10,
            cluster_method='fcm',
            min_count=1,
            window=5,
            vector_size=20,
            workers=4,
            ):

        self.model = ...
        self.word_embedding = ...

        super().__init__(
            algorithm='flsa-e',
            corpus=corpus,
            num_topics=num_topics,
            num_words=num_words,
            cluster_method=cluster_method,
            min_count=min_count,
            window=window,
            vector_size=vector_size,
            workers=workers,
            )

    def get_word_embedding(
            self,
            data,
            vector_size,
            window,
            min_count,
            workers,
            ):
        """
            Method to train a word embedding on the corpus.

            Parameters
                ----------
                    data : list of lists of str
                        The input file used to initialize the model.
                    min_count : int
                        Ignores all words with total frequency lower than this.
                    window : int
                        Maximum distance between the current and predicted word within a sentence.
                    vector_size : int
                        Dimensionality of the word vectors.
                    workers : int
                        Use these many worker threads to train the model
                        ( = faster training with multicore machines).
        """

        self.model = Word2Vec(
            sentences=data,
            vector_size=vector_size,
            window=window,
            min_count=min_count,
            workers=workers,
            )

        return self.model.wv.vectors

    def _get_matrices(
            self,
            ):
        '''
        Method to run after the FLSA_E class has been initialized to obtain the output matrices.

        Returns:
                  - Numpy array: prob_word_given_topic
                  - Numpy array: prob_topic_given_document
        '''
        sparse_document_term_matrix = self._create_sparse_local_term_weights(
            self.corpus,
            self._vocabulary_size,
            self._word_to_index,
            )

        self.word_embedding = self.get_word_embedding(
            data=self.corpus,
            min_count=self.min_count,
            vector_size=self.vector_size,
            window=self.window,
            workers=self.workers,
            )

        partition_matrix = self._create_partition_matrix(
            data=self.word_embedding,
            number_of_clusters=self.num_topics,
            method=self.cluster_method,
            )

        return self._create_probability_matrices(
            algorithm='flsa-e',
            prob_topic_given_word_transpose=partition_matrix,
            local_term_weights=sparse_document_term_matrix,
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2013 Radim Rehurek <radimrehurek@seznam.cz>
# Licensed under the GNU LGPL v2.1 - http://www.gnu.org/licenses/lgpl.html

"""This module contains functions to compute direct confirmation on a pair of words or word subsets."""

import logging

import numpy as np

logger = logging.getLogger(__name__)

# Should be small. Value as suggested in paper http://svn.aksw.org/papers/2015/WSDM_Topic_Evaluation/public.pdf
EPSILON = 1e-12


def log_conditional_probability(segmented_topics, accumulator, with_std=False, with_support=False):
    r"""Calculate the log-conditional-probability measure which is used by coherence measures such as `U_mass`.
    This is defined as :math:`m_{lc}(S_i) = log \frac{P(W', W^{*}) + \epsilon}{P(W^{*})}`.

    Parameters
    ----------
    segmented_topics : list of lists of (int, int)
        Output from the :func:`~gensim.topic_coherence.segmentation.s_one_pre`,
        :func:`~gensim.topic_coherence.segmentation.s_one_one`.
    accumulator : :class:`~gensim.topic_coherence.text_analysis.InvertedIndexAccumulator`
        Word occurrence accumulator from :mod:`gensim.topic_coherence.probability_estimation`.
    with_std : bool, optional
        True to also include standard deviation across topic segment sets in addition to the mean coherence
        for each topic.
    with_support : bool, optional
        True to also include support across topic segments. The support is defined as the number of pairwise
        similarity comparisons were used to compute the overall topic coherence.

    Returns
    -------
    list of float
        Log conditional probabilities measurement for each topic.

    Examples
    --------
    .. sourcecode:: pycon

        >>> from gensim.topic_coherence import direct_confirmation_measure, text_analysis
        >>> from collections import namedtuple
        >>>
        >>> # Create dictionary
        >>> id2token = {1: 'test', 2: 'doc'}
        >>> token2id = {v: k for k, v in id2token.items()}
        >>> dictionary = namedtuple('Dictionary', 'token2id, id2token')(token2id, id2token)
        >>>
        >>> # Initialize segmented topics and accumulator
        >>> segmentation = [[(1, 2)]]
        >>>
        >>> accumulator = text_analysis.InvertedIndexAccumulator({1, 2}, dictionary)
        >>> accumulator._inverted_index = {0: {2, 3, 4}, 1: {3, 5}}
        >>> accumulator._num_docs = 5
        >>>
        >>> # result should be ~ ln(1 / 2) = -0.693147181
        >>> result = direct_confirmation_measure.log_conditional_probability(segmentation, accumulator)[0]

    """
    topic_coherences = []
    num_docs = float(accumulator.num_docs)
    for s_i in segmented_topics:
        segment_sims = []
        for w_prime, w_star in s_i:
            try:
                w_star_count = accumulator[w_star]
                co_occur_count = accumulator[w_prime, w_star]
                m_lc_i = np.log(((co_occur_count / num_docs) + EPSILON) / (w_star_count / num_docs))
            except KeyError:
                m_lc_i = 0.0
            except ZeroDivisionError:
                # if w_star_count==0, it will throw exception of divided by zero
                m_lc_i = 0.0

            segment_sims.append(m_lc_i)

        topic_coherences.append(aggregate_segment_sims(segment_sims, with_std, with_support))

    return topic_coherences


def aggregate_segment_sims(segment_sims, with_std, with_support):
    """Compute various statistics from the segment similarities generated via set pairwise comparisons
    of top-N word lists for a single topic.

    Parameters
    ----------
    segment_sims : iterable of float
        Similarity values to aggregate.
    with_std : bool
        Set to True to include standard deviation.
    with_support : bool
        Set to True to include number of elements in `segment_sims` as a statistic in the results returned.

    Returns
    -------
    (float[, float[, int]])
        Tuple with (mean[, std[, support]]).

    Examples
    ---------
    .. sourcecode:: pycon

        >>> from gensim.topic_coherence import direct_confirmation_measure
        >>>
        >>> segment_sims = [0.2, 0.5, 1., 0.05]
        >>> direct_confirmation_measure.aggregate_segment_sims(segment_sims, True, True)
        (0.4375, 0.36293077852394939, 4)
        >>> direct_confirmation_measure.aggregate_segment_sims(segment_sims, False, False)
        0.4375

    """
    mean = np.mean(segment_sims)
    stats = [mean]
    if with_std:
        stats.append(np.std(segment_sims))
    if with_support:
        stats.append(len(segment_sims))

    return stats[0] if len(stats) == 1 else tuple(stats)


def log_ratio_measure(segmented_topics, accumulator, normalize=False, with_std=False, with_support=False):
    r"""Compute log ratio measure for `segment_topics`.

    Parameters
    ----------
    segmented_topics : list of lists of (int, int)
        Output from the :func:`~gensim.topic_coherence.segmentation.s_one_pre`,
        :func:`~gensim.topic_coherence.segmentation.s_one_one`.
    accumulator : :class:`~gensim.topic_coherence.text_analysis.InvertedIndexAccumulator`
        Word occurrence accumulator from :mod:`gensim.topic_coherence.probability_estimation`.
    normalize : bool, optional
        Details in the "Notes" section.
    with_std : bool, optional
        True to also include standard deviation across topic segment sets in addition to the mean coherence
        for each topic.
    with_support : bool, optional
        True to also include support across topic segments. The support is defined as the number of pairwise
        similarity comparisons were used to compute the overall topic coherence.

    Notes
    -----
    If `normalize=False`:
        Calculate the log-ratio-measure, popularly known as **PMI** which is used by coherence measures such as `c_v`.
        This is defined as :math:`m_{lr}(S_i) = log \frac{P(W', W^{*}) + \epsilon}{P(W') * P(W^{*})}`

    If `normalize=True`:
        Calculate the normalized-log-ratio-measure, popularly knowns as **NPMI**
        which is used by coherence measures such as `c_v`.
        This is defined as :math:`m_{nlr}(S_i) = \frac{m_{lr}(S_i)}{-log(P(W', W^{*}) + \epsilon)}`

    Returns
    -------
    list of float
        Log ratio measurements for each topic.

    Examples
    --------
    .. sourcecode:: pycon

        >>> from gensim.topic_coherence import direct_confirmation_measure, text_analysis
        >>> from collections import namedtuple
        >>>
        >>> # Create dictionary
        >>> id2token = {1: 'test', 2: 'doc'}
        >>> token2id = {v: k for k, v in id2token.items()}
        >>> dictionary = namedtuple('Dictionary', 'token2id, id2token')(token2id, id2token)
        >>>
        >>> # Initialize segmented topics and accumulator
        >>> segmentation = [[(1, 2)]]
        >>>
        >>> accumulator = text_analysis.InvertedIndexAccumulator({1, 2}, dictionary)
        >>> accumulator._inverted_index = {0: {2, 3, 4}, 1: {3, 5}}
        >>> accumulator._num_docs = 5
        >>>
        >>> # result should be ~ ln{(1 / 5) / [(3 / 5) * (2 / 5)]} = -0.182321557
        >>> result = direct_confirmation_measure.log_ratio_measure(segmentation, accumulator)[0]

    """
    topic_coherences = []
    num_docs = float(accumulator.num_docs)
    for s_i in segmented_topics:
        segment_sims = []
        for w_prime, w_star in s_i:
            w_prime_count = accumulator[w_prime]
            w_star_count = accumulator[w_star]
            co_occur_count = accumulator[w_prime, w_star]

            if normalize:
                # For normalized log ratio measure
                numerator = log_ratio_measure([[(w_prime, w_star)]], accumulator)[0]
                co_doc_prob = co_occur_count / num_docs
                m_lr_i = numerator / (-np.log(co_doc_prob + EPSILON))
            else:
                # For log ratio measure without normalization
                numerator = (co_occur_count / num_docs) + EPSILON
                denominator = (w_prime_count / num_docs) * (w_star_count / num_docs)
                m_lr_i = np.log(numerator / denominator)

            segment_sims.append(m_lr_i)

        topic_coherences.append(aggregate_segment_sims(segment_sims, with_std, with_support))

    return topic_coherences



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2010 Radim Rehurek <radimrehurek@seznam.cz>
# Licensed under the GNU LGPL v2.1 - http://www.gnu.org/licenses/lgpl.html

"""
Automated tests for checking transformation algorithms (the models package).
"""

import numpy as np


class TestBaseTopicModel:
    def test_print_topic(self):
        topics = self.model.show_topics(formatted=True)
        for topic_no, topic in topics:
            self.assertTrue(isinstance(topic_no, int))
            self.assertTrue(isinstance(topic, str))

    def test_print_topics(self):
        topics = self.model.print_topics()

        for topic_no, topic in topics:
            self.assertTrue(isinstance(topic_no, int))
            self.assertTrue(isinstance(topic, str))

    def test_show_topic(self):
        topic = self.model.show_topic(1)

        for k, v in topic:
            self.assertTrue(isinstance(k, str))
            self.assertTrue(isinstance(v, (np.floating, float)))

    def test_show_topics(self):
        topics = self.model.show_topics(formatted=False)

        for topic_no, topic in topics:
            self.assertTrue(isinstance(topic_no, int))
            self.assertTrue(isinstance(topic, list))
            for k, v in topic:
                self.assertTrue(isinstance(k, str))
                self.assertTrue(isinstance(v, (np.floating, float)))

    def test_get_topics(self):
        topics = self.model.get_topics()
        vocab_size = len(self.model.id2word)
        for topic in topics:
            self.assertTrue(isinstance(topic, np.ndarray))
            # Note: started moving to np.float32 as default
            # self.assertEqual(topic.dtype, np.float64)
            self.assertEqual(vocab_size, topic.shape[0])
            self.assertAlmostEqual(np.sum(topic), 1.0, 5)
